#__author__ = "magnelPy"

####################
## MODULE IMPORTS ##
####################

import numpy as np
import pandas as pd
from copy import deepcopy

from scipy import optimize
import itertools

####################
## SURROGATE EVAL ##
####################

def y_regression(X,theta):

	htheta=np.dot(X,theta)

	return htheta

###################
## COST FUNCTION ##
###################

#---------------------------------CostFunction -----------------------------
def cost_func(X_tr,Y_tr,theta,SW_grad=False):
	""" Base cost function regression

	Parameters
	----------
	X_tr :	(m,n+1) numpy.ndarray
		n input features for m samples, including unit bias term in first column
	Y_tr :	(m,) numpy.ndarray
		target output for m samples
	theta :	(n+1,) numpy.ndarray
		regression coefficients for the bias term + n features
	SW_grad : Boolean
		return gradient of J wrt theta yes/no

	Returns
	-------
	jtheta1:	scalar
		Base cost regression
	Gjtheta1:	(n+1,) numpy.ndarray
		Gradient of cost with theta
	Reference
	---------
	No external reference

	Prior implementation
	---------
	Ranjit Chaudhary : IPW2020 code : Regression.py

	Examples
	--------
	>>> from magnelPy.ML import regression as reg
	>>> X=np.array([[1,2],[1,5]]) # data points x = 2 and x = 5
	>>> Y=np.array([2,4])
	>>> theta=np.array([1,1])
 	>>> J=reg.cost_func(X,Y,theta)
 	>>> print(J)
 	1.25
	"""
	m=np.shape(X_tr)[0];
	htheta=y_regression(X_tr,theta)
	Y_tr=Y_tr.squeeze()
	Z=np.subtract(htheta,Y_tr)
	# test=pd.DataFrame([htheta,Y_tr,Z],index=['htheta','Y_tr','Z']);print(test.T)
	jtheta1=np.sum(np.square(Z))/(2*m)

	if not SW_grad: return jtheta1
	else:
		Gjtheta1=np.matmul(Z,X_tr)/m
		return jtheta1,Gjtheta1

#--------------------------with Regularization parameter---------------
def cost_funcreg(X_tr,Y_tr,theta,lamda=0,SW_grad=False):
	""" Regularized cost function regression - ridge regression

	Parameters
	----------
	X_tr :	(m,n+1) numpy.ndarray
		n input features for m samples, including unit bias term in first column
	Y_tr :	(m,) numpy.ndarray
		target output for m samples
	theta :	(n+1,) numpy.ndarray
		regression coefficients for the bias term + n features
	lamda : scalar
		regularization coefficient
	SW_grad : Boolean
		return gradient of J wrt theta yes/no

	Returns
	-------
	jtheta:	scalar
		Regularized cost regression
	Gjtheta:	(n+1,) numpy.ndarray
		Gradient of cost with theta

	Reference
	---------
	No external reference

	Prior implementation
	---------
	Ranjit Chaudhary : IPW2020 code : Regression.py

	Examples
	--------
	>>> from magnelPy.ML import regression as reg
	>>> X=np.array([[1,2],[1,5]]) # data points x = 2 and x = 5
	>>> Y=np.array([2,4])
	>>> theta=np.array([1,1])
	>>> lamda = 0.2
 	>>> Jreg=reg.cost_funcreg(X,Y,theta,lambda)
 	>>> print(Jreg)
 	1.30
	"""
	m=np.shape(X_tr)[0]

	if not SW_grad: jtheta1 = cost_func(X_tr,Y_tr,theta)
	else: jtheta1,Gjtheta1 = cost_func(X_tr,Y_tr,theta,True)

	reg_term=(lamda/(2*m))*np.sum(np.square(theta[1:]))
	jtheta=jtheta1+reg_term

	if not SW_grad: return jtheta
	else:
		Greg=lamda*theta/m; Greg[0]=0
		Gjtheta=Gjtheta1+Greg
		return jtheta, Gjtheta

###########################
## OPTIMIZATION FUNCTION ##
###########################

#----------------------------Gradient Descent Algorithm-----------------------
def Gradient_descentalg(X_tr,Y_tr,lamda,theta,alpha,cost_tol=10**(-5),maxiter=10**4,SW_fullOutput=False,SW_debug=False):
	""" Train regularized regression - Gradient descent algorithm

	Parameters
	----------
	X_tr :	(m,n+1) numpy.ndarray
		n input features for m samples, including unit bias term in first column
	Y_tr :	(m,) numpy.ndarray
		target output for m samples
	lamda : scalar
		regularization coefficient
	theta :	(n+1,) numpy.ndarray
		regression coefficients for the bias term + n features
	alpha : scalar
		step-size update gradient descent algorithm
	cost_tol : scalar
		convergence criterium : cost tolerance in regularized cost calculation  
	maxiter : integer
		convergence criterium : maximum number of iterations
	SW_fullOutput : Boolean
		switch for calculation step logging and return

	Returns
	-------
	theta:	(n+1,) numpy.ndarray
		regression coefficients for the bias term + n features

	if SW_fullOutput :
	fullOutput :	Dict
		Dict with logged calculations

	Reference
	---------
	No external reference

	Prior implementation
	---------
	Ranjit Chaudhary : IPW2020 code : Regression.py

	Examples
	--------
	>>> from magnelPy.ML import regression as reg
	>>> X=np.array([[1,2],[1,5]]) # data points x = 2 and x = 5
	>>> Y=np.array([2,4])
	>>> theta=np.array([1,1])
	>>> lamda = 0.
	>>> alpha=0.05
 	>>> theta,fullOutput=reg.Gradient_descentalg(X,Y,lamda,theta,alpha,cost_tol=10**-7,maxiter=10,SW_fullOutput=True)
 	>>> print(fullOutput['thetalist'])
				   0         1
		0   1.000000  1.000000
		1   0.925000  0.700000
		2   0.906250  0.630625
		3   0.900578  0.614828
		4   0.897954  0.611477
		5   0.896048  0.611014
		6   0.894318  0.611220
		7   0.892639  0.611580
		8   0.890980  0.611973
		9   0.889336  0.612371
		10  0.887704  0.612768
	# note: analytical result : theta_0=theta_1=2./3
	"""
	# squeeze for dimension issue
	Y_tr=Y_tr.squeeze()

	# initial cost and approximation
	cost = cost_funcreg(X_tr,Y_tr,theta,lamda)
	htheta=y_regression(X_tr,theta) # current regression approximation
	# calc parameters
	m,n=np.shape(X_tr); 
	lamda_v=lamda*np.ones(n); lamda_v[0]=0.
	# gradient descent iteration setup
	iterate=True; niter=0
	if SW_fullOutput: 
		fullOutput={} # Dict of calc info
		Jlist=[cost]; thetalist=pd.DataFrame(theta,columns=[0]); Yapproxlist=pd.DataFrame(htheta,columns=[0])
	# gradient descent iteration
	while iterate:
		niter+=1
		## matrix implementation
		if not SW_debug:
			A=alpha/m*np.matmul(np.transpose(X_tr),y_regression(X_tr,theta)-Y_tr) # update matrix, excl. regression
			B=1-alpha*lamda_v/m
			theta=np.multiply(theta,B)-A
		if SW_debug:
		## original step-wise - maintained for demonstration purposes : set SW_debug == True
			for i in np.arange(len(theta)):
				redn= alpha/m*np.sum(np.multiply(htheta-Y_tr,X_tr[:,i]))
				if i==0:
					theta[i]=theta[i]-redn
				else:
					theta[i]=theta[i]*(1-alpha*lamda/m)-redn
		new_cost = cost_funcreg(X_tr,Y_tr,theta,lamda)
		htheta=y_regression(X_tr,theta) # current regression approximation
		if SW_fullOutput: Jlist.append(new_cost); thetalist[niter]=theta; Yapproxlist[niter]=htheta
		if abs(new_cost-cost)<=cost_tol: iterate=False
		if niter==maxiter: iterate=False
		cost = new_cost
	if SW_fullOutput:
		fullOutput['Jlist']=Jlist; fullOutput['thetalist']=thetalist.T; fullOutput['Yapproxlist']=Yapproxlist.T
		return theta,fullOutput
	else: return theta

def train_regularizedRegression(X_tr,Y_tr,lamda,theta0,SW_optimObject=False):
	"""  Train regularized regression - apply fmin_cg

	WIP - TO BE COMPLETED

	SW_optimObject : Boolean
		switch for return of optimObject; if False only optim values are returned

	"""

	## package arguments
	args=(X_tr,Y_tr,lamda)

	## optimization options
		# maxiter 	: int 		- Maximum number of iterations to perform. Default is 200 * len(x0)
		# disp		: Boolean	- If True, return a convergence message, followed by xopt
		# gtol		: float		- Stop when the norm of the gradient is less than gtol
	opts={'maxiter' : None,
         'disp' : False,
         'gtol' : 1e-5}

	## optimize	
	res=optimize.minimize(packed_cost_funcreg, theta0, jac=True, args=args, method='CG',options=opts)

	if not SW_optimObject: return res.x
	else: return res

##################
## FEATURE PREP ##
##################

def poly_model(Xmain,k,interaction=True):
	""" Polynomial expansion of feature set + bias addition

	Parameters
	----------
	Xmain :	(m,n) pd.DataFrame
		n input features for m samples
	k	:	scalar
		degree of feature polynomial
	interaction : Boolean
		include interation terms yes/no

	Returns
	-------
	X	:	(m,N +1) numpy.ndarray
		data for kth order polynomial regression; N ifo interaction yes/no

	Reference
	---------
	No external reference

	Prior implementation
	---------
	Ranjit Chaudhary : IPW2020 code : Regression.py

	Future development
	------------------
	List 'XX' as 'X2' etc for higher order features

	Examples
	--------
	WIP
	>>> from magnelPy.ML import regression as reg
	>>> X=np.array([[1,2],[1,5]]) # data points x = 2 and x = 5
	>>> Y=np.array([2,4])
	>>> theta=np.array([1,1])
	"""

	## setup
	X=deepcopy(Xmain)
	columns=list(X.columns); n=len(X.columns); m=len(X.index) # number of features

	## feature mapping
	comb_list = num_pol_terms(k,n,interaction)
	out=pd.DataFrame()

	for tup in comb_list:
		if len(tup)==1:
			header=columns[tup[0]-1]
			out[header]=X[header]
		else:
			val=np.ones(m); header=[]
			for j in tup:
				feature=columns[j-1]
				val*=X[feature]
				header+=feature
			out["".join(header)]=val

	return out

def addBias(data):

	## add bias
	columns=list(data.columns)
	columns=['bias']+columns
	data['bias']=1

	return data[columns]

#----------------------------Data (feature) scaling-----------------------
def DataScale(data,scale_param=None):
	""" Data feature / output scaling

	Parameters
	----------
	data :	np.ndarray or pd.DataFrame
		m realizations in rows; n different features in columns (or single feature)

	Returns
	-------
	x	:	np.ndarray
		m x n array of scaled features
	WIP

	Reference
	---------
	No external reference

	Examples
	--------
	WIP
	>>> from magnelPy.ML import regression as reg
	>>> X=np.array([[1,2],[1,5]]) # data points x = 2 and x = 5

	"""

	if not isinstance(scale_param,pd.DataFrame):
		## scaling from scratch
		# transform pd objects to np.ndarray for consistency standard deviation calc
		if isinstance(data,pd.DataFrame):
			SW_df=True; index=data.index; column=data.columns
			data=data.values
		else: SW_df=False
		# scale
		x,m,std=scale(data)
		# output - differentiation for output naming columns (overkill)
		if SW_df:
			x=pd.DataFrame(x,index=index,columns=column)
			scale_param=pd.DataFrame(np.array([m,std]),index=['mean','std'],columns=column)
		else:
			x=pd.DataFrame(x); scale_param=pd.DataFrame(np.array([m,std]),index=['mean','std'])
		return x,scale_param
	else:
		## scaling with pre-specified scaling parameters
		x=(data-scale_param.loc['mean'])/scale_param.loc['std']
		return x



##################
## AUX FUNCTION ##
##################

def scale(x,SW_sample_std=True):
	""" scales data as ksi = (x-x.mean())/x.std()

	Parameters
	----------
	x :	np.ndarray
		m observations (rows) x n features (columns)
	SW_sample_std : Boolean
		True for sample standard deviation : conform Matlab Coursera

	Note
	----------
	x.std() is defined differently for np.ndarray and pd.Series (division sqrt.(m) vs. sqrt(m-1))
	"""
	m=x.shape[0]
	mean=x.mean(axis=0)
	std=x.std(axis=0)
	if SW_sample_std: std=std*np.sqrt(m)/np.sqrt(m-1)
	return (x-mean)/std,mean,std

def packed_cost_funcreg(theta,*args):

	## unpacking args
	(X_tr,Y_tr,lamda)=args

	## call regularized cost function
	jtheta, Gjtheta=cost_funcreg(X_tr,Y_tr,theta,lamda,SW_grad=True)

	return jtheta,Gjtheta

def num_pol_terms(PolDegree,nVar,interaction=True):
	""" Combination list for polynomial feature mapping

	Parameters
	----------
	PolDegree 	:	int
		degree of polynomial
	nVar		:	int
		number of features
	interaction	:	Boolean
		feature interaction yes/no

	Returns
	-------
	Combination :	list of tuples
		list of tuples with feature combinations (featurs 1 to n)

	Reference
	---------
	No external reference

	Prior implementation
	---------
	Ranjit Chaudhary : IPW2020 code : Regression.py (full copy)

	Examples
	--------
	>>> from magnelPy.ML import regression as reg
	>>> k=2; n=3; SW_interaction=False
	>>> comb=reg.num_pol_terms(k,n,SW_interaction); print(comb)
	[(1,), (1, 1), (2,), (2, 2), (3,), (3, 3)]
	>>> SW_interaction=True
	>>> comb=reg.num_pol_terms(k,n,SW_interaction); print(comb)
	[(1,), (1, 1), (1, 2), (1, 3), (2,), (2, 2), (2, 3), (3,), (3, 3)]

	"""
	var = np.arange(nVar)+1
	for num in np.arange(PolDegree):
		if num==0:
			combination = list(itertools.combinations_with_replacement(var,num+1))
		else:
			comb= list(itertools.combinations_with_replacement(var,num+1))
			combination=combination+comb
	combination.sort(); ncombo=[]
	if not interaction:
		for tup in combination:
			if len(tup)>=1 and all(np.in1d(tup, tup[0])):
				ncombo.append(tup)
		combination = ncombo
	return combination

#################
## FUNCTION RC ##
#################

#---------Performs polynomial expansions and writes in csv file to make easy read during optimization
def Data_Scaling(LearningPath,Ypath,PolDegree,interaction=True):
	wbx=load_workbook(LearningPath); sheets = wbx.sheetnames
	data1=pd.ExcelFile(LearningPath); data2=pd.ExcelFile(Ypath)
	Xmain=data1.parse("X_tr"); Ymain=data2.parse("X_tr")
	if not 'X_cv'in sheets:
		ntotal=np.shape(Xmain.values)[0]
		num_tr=int(0.8*ntotal); num_cv=int(0.15*ntotal); num_test=int(0.05*ntotal)
		X=poly_model(Xmain.values,PolDegree,interaction);Y=Ymain.values[:,1]
		X_tr=X[:num_tr,:];X_cv=X[num_tr:num_tr+num_cv,:]
		X_test=X[num_tr+num_cv:num_tr+num_cv+num_test,:]
		Y_tr=Y[:num_tr];Y_cv=Y[num_tr:num_tr+num_cv]
		Y_test=Y[num_tr+num_cv:num_tr+num_cv+num_test]
	else:
		cvx=data1.parse("X_cv"); cvy=data2.parse("X_cv")
		X_tr=poly_model(Xmain.values,PolDegree,interaction); Y_tr=Ymain.values[:,1]
		splitnum=np.shape(cvx.values)[0]
		X_cv=poly_model(cvx.values,PolDegree,interaction); Y_cv=cvy.values[:,1]
		X_test=X_cv[:int(splitnum*0.5),:];Y_test=Y_cv[:int(splitnum*0.5)]
		X_cv=X_cv[int(splitnum*0.5):,:];Y_cv=Y_cv[int(splitnum*0.5):]

	meanX_tr = np.mean(X_tr,axis = 0);stdX_tr = np.std(X_tr,axis = 0)
	# Scaling the input variables using sklearn library in python
	scaler = StandardScaler()
	X_tr[:,1:] =scaler.fit_transform(X_tr[:,1:]); X_cv[:,1:] =scaler.transform(X_cv[:,1:])
	meanY_tr = np.mean(Y_tr);stdY_tr = np.std(Y_tr)
	Y_tr=(Y_tr-meanY_tr)/stdY_tr;Y_cv=(Y_cv-meanY_tr)/stdY_tr;
	meanX_tr=np.append(meanX_tr,meanY_tr);stdX_tr=np.append(stdX_tr,stdY_tr)
	return X_tr,X_cv,X_test,Y_tr,Y_cv,Y_test,meanX_tr,stdX_tr


# Fits the X-data into the polynomial expression for all the training sets
def poly_model_RC(Xmain,PolDegree,interaction=True):
	comb_list = num_pol_terms(PolDegree,np.shape(Xmain)[1]-1,interaction)
	X=np.zeros((np.shape(Xmain)[0],len(comb_list)+1), dtype=float)
	X[:,0]=np.ones(np.shape(Xmain)[0])
	for ncombo in np.arange(len(comb_list)):
		if len(comb_list[ncombo])==1:
			X[:,ncombo+1]=Xmain[:,int(comb_list[ncombo][0])]
		else:
			for nlocal in np.arange(len(comb_list[ncombo])):
				if nlocal==0:
					X[:,ncombo+1]=Xmain[:,int(comb_list[ncombo][nlocal])]
				else:
					X[:,ncombo+1]=X[:,ncombo+1]*Xmain[:,int(comb_list[ncombo][nlocal])]
	return X

#########################
## STAND ALONE - DEBUG ##
#########################

## Test Function ##
###################

def Test(sentence='Default sentence'):
	""" Print sentence to screen - Test module activation

	Parameters
	----------
	sentence :	str
		input sentence

	Returns
	-------
	None

	Returns
	-------
	None

	Reference
	---------
	None

	Examples
	--------
	>>> import magnelPy.ML as ML
	>>> ML.regression.Test()
	Default sentence
 	>>> ML.regression.Test("Another test")
 	Another test
	"""
	print(sentence)

if __name__ == "__main__":

	Test()
	Test('Another test')
	help(Test)


import numpy as np 
import matplotlib.pyplot as plt 
import _pickle as pickle 
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter, AutoMinorLocator)
import matplotlib as mpl
from magnelPy import statFunc

#####################################################################
# First version, I will put detailed comments soon - Balsa
#################################################################
def oneplot(plot_dict, fig):

	x=plot_dict['x']
	y=plot_dict['y']
	xlim=plot_dict["xlim"]
	ylim=plot_dict["ylim"]
	xscale=plot_dict["xscale"]
	yscale=plot_dict["yscale"]
	label=plot_dict["label"]
	linestyle=plot_dict["linestyle"]
	color=plot_dict["color"]
	marker=plot_dict["marker"]
	#dashes=plot_dict["dashes"]
	try:
		ax = plt.gca()
	except:
		ax = fig.add_subplot(1, 1, 1)

	plot_line=ax.plot(x,y,label=label,linestyle=linestyle,color=color, marker=marker)#,dashes=dashes
	leftx, rightx = plt.xlim()
	# if leftx > xlim[0]: # RVC : issue as this overwrites the command
	leftx=xlim[0]
	# elif rightx<xlim[1]:
	rightx=xlim[1]
	plt.xlim((leftx, rightx))
	# lefty, righty = plt.ylim()
	
	# # if lefty > ylim[0]:
	# lefty=ylim[0]
		
		
	# # if righty < ylim[1]:
	# righty=ylim[1]
	
		
	# plt.ylim((lefty, righty))
	plt.xscale(xscale)
	plt.yscale(yscale)
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))

	#ax.yaxis.set_minor_locator(MultipleLocator(5))
	plt.legend(handlelength=4,framealpha=1,fontsize=12)
	#ax.yaxis.set_major_formatter(FormatStrFormatter('%.4f'))
	ax.tick_params(axis='both',pad=10,labelsize=12, length=5) # mod conform legend and axis titles
	#ax.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter('{x:,g}'))

	return plot_line
def create_plot_dictionary(x,y, xlim=0, ylim=0, xscale=0, yscale=0,linestyle=0, label=0, color=0,marker=0):
	
	if xlim == 0:
		xlength=np.max(x)-np.min(x)
		xlim=(np.min(x)-0.1*xlength,np.max(x)+0.1*xlength)
	if ylim == 0:
		ylength=np.max(y)-np.min(y)
		ylim=(np.min(y)-0.1*ylength,np.max(y)+0.1*ylength)
	if xscale == 0:
		xscale='linear'
	if yscale == 0:
		yscale='linear'
	if label == 0:
		label=''
	if color == 0:
		color='black'
	if marker == 0:
		marker='None'
	if linestyle == 0:
		linestyle='-'




	plot_dict={
	"x":x,
	"y":y,
	"xlim":xlim,
	"ylim":ylim,
	"xscale":xscale,
	"yscale":yscale,
	"label":label,
	"linestyle":linestyle,
	"color":color,
	"marker":marker
	
	}
	return plot_dict
	
	
def color_list(num):
	color_list=['black', 'r', "blue",'green', 'orange',  'm', 'c', 'grey', 'brown','pink','olive']
	color=color_list[num]
	return color

def marker_list(num):
	marker_list=["o","x","s","+","^","D","."]
	marker=marker_list[num]
	return marker

def linestyle_list(num):
	#Dodaj jos linije crtica pomocu recnika preko prve reci
	linestyle_list = ['-', (1,(1,1)),(0,(3,2,1,2)),(0,(5,5)),(0,(5,10)),(0,(3,2,1,2,1,2)) ] 
	linestyle=linestyle_list[num]
	return linestyle

def ploty_cdf(data, names=0,figname=0,BW=False,fig=0,linestyle=0,cdftype='both'):
	
	#################################################################
	#CDF plotting function
	#	data - 		input array where a seperate CDF will be drawn for every column
	#	names - 	list of names for every column
	#	figename - 	name of the pickle file where the figure will be saved (if = 0 no saveing)
	#
	#	type - 		type of visulization : 'cdf', 'ccdf', 'both'
	#################################################################
	try:
		data[0,0]
	except:
		data=data.reshape((len(data), 1))
	cdf=data*0
	for num in np.arange(len(data[0,:])):
		cdf[:,num]=np.sort(data[:,num])
	if names == 0:
		names=[None]*len(cdf[0,:]) 
	if linestyle == 0:
		linestyle='-'
	if fig == 0:
		fig=plt.figure(figsize=(8.0,6.0))
	plot_dicts=[]
	n=len(cdf[:,num])
	for num in np.arange(len(cdf[0,:])):
		if BW: # RVC : linestyle is overwritten
			if cdftype=='both' or cdftype=='cdf': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=np.arange(0+1/n/2,1,1/len(cdf[:,num])),label=names[num],yscale='Log', linestyle=linestyle))
			if cdftype=='both': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0+1/n/2,1,1/len(cdf[:,num])),yscale='Log',linestyle=linestyle))
			if cdftype=='ccdf': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0+1/n/2,1,1/len(cdf[:,num])),label=names[num],yscale='Log',linestyle=linestyle))
		else:
			if cdftype=='both' or cdftype=='cdf': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=np.arange(0+1/n/2,1,1/len(cdf[:,num])),color=color_list(num),label=names[num],yscale='Log',linestyle=linestyle ))
			if cdftype=='both': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0+1/n/2,1,1/len(cdf[:,num])),color=color_list(num),yscale='Log',linestyle=linestyle))
			if cdftype=='ccdf': plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0+1/n/2,1,1/len(cdf[:,num])),color=color_list(num),label=names[num],yscale='Log',linestyle=linestyle))
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)
	# plt.ylim((1/len(cdf[:,0]/2),1))
	fig.show()
	

	if figname !=0:
		#print(figname)
		pickle.dump(fig,open(figname,'wb'))
	return fig

def ploty_pdf(data, names=0,nbins=50,figname=0,fig=0,linestyle=0):
	
	#################################################################
	#PDF plotting function
	#	data - 		input array where a seperate PDF will be drawn for every column
	#	names - 	list of names for every column
	#	nbins - 	total number of bins for the whole graph (each PDF will have the same bin width)
	#	figename - 	name of the pickle file where the figure will be saved (if = 0 no saveing)
	#
	#################################################################
	try:
		data[0,0]
	except:
		data=data.reshape((len(data), 1))

	my_bins=np.arange(0.9*np.min(data),1.1*np.max(data),(1.1*np.max(data)-0.9*np.min(data))/nbins)
	bwidth=(1.1*np.max(data)-0.9*np.min(data))/nbins
	Histy=np.zeros((len(my_bins),len(data[0,:])))
	Histx=np.zeros((len(my_bins)+1,len(data[0,:])))

	if names == 0:
		names=['']*len(data[0,:])
	if linestyle == 0:
		linestyle='-'
	if fig == 0:
		fig=plt.figure(figsize=(18,10))
	plot_dicts=[]
	for num in np.arange(len(data[0,:])):
		Histy[1:,num],Histx[1:,num]=np.histogram(data[:,num],bins=my_bins)
		Histx[0,num]=Histx[1,num]
		plot_dicts.append(create_plot_dictionary(x=(Histx[:-1,num]+Histx[1:,num])/2,y=Histy[:,num]/len(data[:,num])/bwidth,color=color_list(num),label=names[num],linestyle=linestyle))
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)
	if figname !=0:
		pickle.dump(fig,open(figname,'wb'))
	lefty, righty = plt.ylim()
	plt.ylim((0, righty))

	fig.show()
	return fig

def ploty_cdf_approx(m,s,fig,dist_type):
	
	leftx, rightx = plt.xlim()
	cdfx=np.arange(leftx,rightx,(rightx-leftx)/100)
	plot_dicts=[]
	for num in np.arange(len(m)):

		if dist_type == "N":
			cdfy=statFunc.F_Normal(cdfx,m[num],s[num])
		elif dist_type == 'LN':
			cdfy=statFunc.F_Lognormal(cdfx,m[num],s[num])
		elif dist_type == 'Gamma':
			cdfy=statFunc.F_Gamma(cdfx,m[num],s[num])	
		elif dist_type == 'Gumbel':
			cdfy=statFunc.F_Gumbel(cdfx,m[num],s[num])
		plot_dicts.append(create_plot_dictionary(x=cdfx,y=cdfy,color=color_list(num),yscale='Log' ,linestyle=':'))
		plot_dicts.append(create_plot_dictionary(x=cdfx,y=1-cdfy,color=color_list(num),yscale='Log',linestyle=':'))
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)

	return fig

def ploty_pdf_approx(m,s,fig,dist_type):
	
	leftx, rightx = plt.xlim()
	pdfx=np.arange(leftx,rightx,(rightx-leftx)/100)
	plot_dicts=[]
	for num in np.arange(len(m)):

		if dist_type == "N":
			pdfy=statFunc.f_Normal(pdfx,m[num],s[num])
		elif dist_type == 'LN':
			pdfy=statFunc.f_Lognormal(pdfx,m[num],s[num])
		elif dist_type == 'Gamma':
			pdfy=statFunc.f_Gamma(pdfx,m[num],s[num])	
		elif dist_type == 'Gumbel':
			pdfy=statFunc.f_Gumbel(pdfx,m[num],s[num])
		plot_dicts.append(create_plot_dictionary(x=pdfx,y=pdfy,color=color_list(num) ,linestyle=':'))
		
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)
	lefty, righty = plt.ylim()
	plt.ylim((0, righty))

	return fig

def ploty_cdf_CONFAB(data, names=0,figname=0,BW=False,fig=0):
	
	#################################################################
	#CDF plotting function
	#	data - 		input array where a seperate CDF will be drawn for every column
	#	names - 	list of names for every column
	#	figename - 	name of the pickle file where the figure will be saved (if = 0 no saveing)
	#
	#################################################################
	try:
		data[0,0]
	except:
		data=data.reshape((len(data), 1))
	cdf=data*0
	for num in np.arange(len(data[0,:])):
		cdf[:,num]=np.sort(data[:,num])
	if names ==0:
		names=[None]*len(cdf[0,:])
	if fig == 0:
		fig=plt.figure(figsize=(12,9))
	plot_dicts=[]
	for num in np.arange(len(cdf[0,:])):
		if BW:
			plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=np.arange(0,1,1/len(cdf[:,num])),label=names[num],yscale='Log', linestyle=linestyle_list(num)))
			plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0,1,1/len(cdf[:,num])),yscale='Log',linestyle=linestyle_list(num)))
		else:
			plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=np.arange(0,1,1/len(cdf[:,num])),color=color_list(int(num/3)),linestyle=linestyle_list(num % 3),label=names[num],yscale='Log' ))
			plot_dicts.append(create_plot_dictionary(x=cdf[:,num],y=1-np.arange(0,1,1/len(cdf[:,num])),color=color_list(int(num/3)),linestyle=linestyle_list(num % 3),yscale='Log'))
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)
	plt.ylim((1/len(cdf[:,0]),1))
	#fig.show()
	

	if figname !=0:
		print(figname)
		pickle.dump(fig,open(figname,'wb'))
	return fig
def ploty_pdf_CONFAB(data, names=0,nbins=50,figname=0,fig=0,color=0):
	
	#################################################################
	#PDF plotting function
	#	data - 		input array where a seperate PDF will be drawn for every column
	#	names - 	list of names for every column
	#	nbins - 	total number of bins for the whole graph (each PDF will have the same bin width)
	#	figename - 	name of the pickle file where the figure will be saved (if = 0 no saveing)
	#
	#################################################################
	try:
		data[0,0]
	except:
		data=data.reshape((len(data), 1))

	my_bins=np.arange(0.9*np.min(data),1.1*np.max(data),(1.1*np.max(data)-0.9*np.min(data))/nbins)
	bwidth=(1.1*np.max(data)-0.9*np.min(data))/nbins
	Histy=np.zeros((len(my_bins),len(data[0,:])))
	Histx=np.zeros((len(my_bins)+1,len(data[0,:])))

	if names == 0:
		names=['']*len(data[0,:])
	if fig == 0:
		fig=plt.figure(figsize=(12,9))
	plot_dicts=[]
	for num in np.arange(len(data[0,:])):
		Histy[1:,num],Histx[1:,num]=np.histogram(data[:,num],bins=my_bins)
		Histx[0,num]=Histx[1,num]
		plot_dicts.append(create_plot_dictionary(x=(Histx[:-1,num]+Histx[1:,num])/2,y=Histy[:,num]/len(data[:,num])/bwidth,color=color_list(color),linestyle=linestyle_list(num % 3),label=names[num]))
	for num in np.arange(len(plot_dicts)):
		oneplot(plot_dicts[num],fig)
	if figname !=0:
		pickle.dump(fig,open(figname,'wb'))
	lefty, righty = plt.ylim()
	plt.ylim((0, righty))
	fig.show()
	return fig




if __name__ == "__main__":
	
	print('Test')
	Pn_input =np.load("Pn8.npy")
	names=[r'$E_{Confab}$',r'$E_{H&S}$',r'$E_{Iqbal}$',r'$E_{Confab}$',r'$E_{H&S}$',r'$E_{Iqbal}$',r'$E_{Confab}$',r'$E_{H&S}$']
	#ploty_pdf(Pn_input[:,:3],names)
	ploty_cdf(Pn_input[:,:6],names,BW=True,figname='blabla')
	plt.show()

	
	# fig=plt.figure()

	# plot1=create_plot_dictornary(x=np.sort(Pn_input[:,0]),y=np.arange(0,1,1/len(Pn_input[:,0])),label='prvi',color='blue',yscale='log')
	# plot2=create_plot_dictornary(x=np.sort(Pn_input[:,1]),y=np.arange(0,1,1/len(Pn_input[:,1])),label='drugi',color='red',yscale='log')
	# plot1c=create_plot_dictornary(x=np.sort(Pn_input[:,0]),y=1-np.arange(0,1,1/len(Pn_input[:,0])),color='blue',yscale='log')
	# plot2c=create_plot_dictornary(x=np.sort(Pn_input[:,1]),y=1-np.arange(0,1,1/len(Pn_input[:,1])),color='red',yscale='log')

	# oneplot(plot1,fig)
	# oneplot(plot2,fig)
	# oneplot(plot1c,fig)
	# oneplot(plot2c,fig)
	# plt.xlim((0,1.1*np.max(Pn_input[:,0])))

	# plt.draw()
	# plt.show()


"""Capacity\\EC3
Notes
-----
EN 1993-1-2:2005 capacity evaluations
note: EN 1993-1-2:2005 further referenced as EC
"""

#############
## IMPORTS ##
#############

import numpy as np
from magnelPy.SFE.MaterialProperties import Steel as steel


############
## MODULE ##
############

def steelcol_thetacrit(Hfi,NEdfi,Icol,Acol,fy,E=200000):
	""" Steel column critical temperature, considering capacity according to section 4.2.3.2 EC;

	Parameters
	----------	
	Hfi   : float
		[-], buckling length for fire design - see Figure 4.1 EC
	NEdfi : float
		[N], fire design load
	Icol  : float
		[mm4], cross-section moment of inertia - consider weak axis / strong axis ifo buckling length 
	Acol  : float
		[mm2], cross-section area
	fy	: float
		[N/mm2], steel yield stress in normal design conditions
	E	 : float
		[N/mm2], steel modulus of elasticity in normal design conditions

	Returns
	-------
	theta_crit : float
		 [C], critical temperature for axially loaded steel column
	
	Example use
	-----------
	from magnelPy.SFE.Capacity import EC3
	Hfi=3500 # [mm]
	Efi=9587*10**3 # [N]
	Icol=8.27*10**8 # [mm4]
	Acol=70.2*10**3 # [mm2] 
	fy=235 # [N/mm2]
	E=200000 # [N/mm2]
	theta_crit=EC3.steelcol_thetacrit(Hfi,Efi,Icol,Acol,fy,E) # [C]; critical temperature
	print("The critical temperature is {0:.0f} degrees C.".format(theta_crit))
	>>> The critical temperature is 508 degrees C.
	"""

	## aux
	tem=np.arange(100,1200,1)

	## column capacity ifo pre-listed critical temperature
	NRdfi=steelcol_capacityGivenTemperature(tem,Hfi,Icol,Acol,fy,E)

	## critical temperature
	theta_crit=np.interp(-NEdfi,-NRdfi,tem) # interpolate on negative values, since np.interp requires x-values to be increasing

	return theta_crit

def steelcol_capacityGivenTemperature(theta,Hfi,Icol,Acol,fy,E=200000,SW_elementwise=False):
	""" Steel column capacity for given critical temperature, according to section 4.2.3.2 EC;

	Parameters
	----------	
	theta : float or np.array (n,) or (m,)
		[C], uniform section temperature
	Hfi : float
		[-], buckling length for fire design - see Figure 4.1 EC
	Icol  : float
		[mm4], cross-section moment of inertia - consider weak axis / strong axis ifo buckling length 
	Acol  : float
		[mm2], cross-section area
	fy	: float or np.array (n,)
		[N/mm2], steel yield stress in normal design conditions
	E	 : float
		[N/mm2], steel modulus of elasticity in normal design conditions
	SW_elementwise : Boolean
		Combine array element-wise; requires arrays of same shape - now assumed (n,)

	Returns
	-------
	theta_crit : float
		 [C], critical temperature for axially loaded steel column
	
	Example use
	-----------
	from magnelPy.SFE.Capacity import EC3
	import numpy as np
	Hfi=3500 # [mm]
	Icol=8.27*10**8 # [mm4]
	Acol=70.2*10**3 # [mm2] 
	fy=235 # [N/mm2]
	E=200000 # [N/mm2]
	temperature=np.array([20,100,300,400,500,600,700])
	NRdfi=EC3.steelcol_capacityGivenTemperature(temperature,Hfi,Icol,Acol,fy,E) # [C]; critical temperature
	for i,tem in enumerate(temperature):
		print("At {0:.1f} degrees C, the design capacity is {1:.0f} kN.".format(tem,NRdfi[i]*10**-3))
	>>> At 20.0 degrees C, the design capacity is 13158 kN.
	>>> At 100.0 degrees C, the design capacity is 13158 kN.
	>>> At 300.0 degrees C, the design capacity is 12783 kN.
	>>> At 400.0 degrees C, the design capacity is 12538 kN.
	>>> At 500.0 degrees C, the design capacity is 9916 kN.
	>>> At 600.0 degrees C, the design capacity is 5839 kN.
	>>> At 700.0 degrees C, the design capacity is 2785 kN.
	"""

	## hardcoded values
	Ymfi = 1.0 # global resistance factor; recommended EC ?

	## broadcast handling

	if not SW_elementwise:
		if not isinstance(theta,np.ndarray):
			theta=np.array([theta])
		if not isinstance(fy,np.ndarray):
			fy=np.array([fy])
		fy=fy[:,np.newaxis]
		theta=theta[np.newaxis,:]

	## input handling
	rcol=np.sqrt(Icol/Acol) # [mm]; radius of gyration 

	## buckling resistance calculation according to 4.2.3.2 EC
	lambda_i = np.sqrt(E/fy)*np.pi
	Lambda_ = (Hfi/rcol)*1/lambda_i
	alpha = 0.65*np.sqrt(235/fy)
	Lambda_theta = Lambda_*np.sqrt(steel.ky_EC3(theta)/steel.kE_EC3(theta))
	fi_theta = 0.5*(1+alpha*Lambda_theta+Lambda_theta**2)
	chi_fi = 1/(fi_theta+np.sqrt(fi_theta**2-Lambda_theta**2))
	NRdfi=chi_fi*Acol*steel.ky_EC3(theta)*fy/Ymfi

	return np.squeeze(NRdfi) # squeeze to remove 1D array


def Temperature_LumpedMass_Unprotected(Tg,time,AV,hc,Ksh):
	""" Unprotected steel section temperature, according to section 4.2.5.1 EC;

	Parameters
	----------	
	Tg : np.array(n,m)
		[C], adiabatic surface temperature (gas temperature)
		with n the number of fire curves, and m the number of timesteps
	time : np.array(m,)
		[s], timesteps for the calculation
	hc : float
		[W/(m2K)], convection coefficient
	Av  : float
		[1/m], section factor for the unprotected section
	Ksh  : float
		[-], correction factor for the shadow effect

	Returns
	-------
	Ts : np.array(n,m)
	 [C],  temperature for unprotected steel section

	Example use
	-----------
	import numpy as np
	import magnelPy.SFE as sfe
	from magnelPy.SFE.MaterialProperties import Steel as steel
	from magnelPy.SFE.Capacity import EC3 as EC3
	# parameters steel temperature calculation
	AV = 107 # [1/m]; specific surface section
	AVb = 80 # [1/m]; specific surface boxed
	Ksh = 0.9*AVb/AV # [-]; shadow factor
	hc=35 # [W/m2K]; convection coefficient
	# fire curve definition
	O = 0.04; qf=780 # parameters EPFC
	timelist=np.arange(0,360*60+0.5,0.5) #[s]; timesteps
	Tg=np.array(sfe.FireCurve.EuroCodeParametric(timelist/60, O, qf))
	Ts=EC3.Temperature_LumpedMass_Unprotected(Tg,timelist,AV,hc,Ksh)
	times=[0,60,120,180,240,360] # [min]
	steps=np.array(times)*60*2
	for i,timepoint in enumerate(times):
		print("At {0:.0f} min, the temperature is {1:.0f} degrees C.".format(timepoint,Ts[steps[i]]))
	>>> At 0 min, the temperature is 20 degrees C.
	>>> At 60 min, the temperature is 861 degrees C.
	>>> At 120 min, the temperature is 709 degrees C.
	>>> At 180 min, the temperature is 356 degrees C.
	>>> At 240 min, the temperature is 72 degrees C.
	>>> At 360 min, the temperature is 20 degrees C.
	"""
	
	# hardcoded properties
	emissivity=0.7
	rho=steel.density_EC3() # [kg/m3] steel density (constant)
	
	# initialization
	try: 
		(n,m)=np.shape(Tg)
	except:
		Tg=Tg[np.newaxis,:]
		(n,m)=np.shape(Tg)
	Ts = np.zeros((n,m))
	newT=20*np.ones(n) # [C] starting temperature

	# forward integration
	for i in np.arange(m-1):
		Ts[:,i]=newT # assign temperature
		# steel property at elevated temperature
		cp=steel.cp_EC3(Ts[:,i]) # [J/kgK]; specific heat
		# heat flux
		heat_flux = (hc*(Tg[:,i]-Ts[:,i])+emissivity*5.67*10**-8*((Tg[:,i]+273.15)**4-(Ts[:,i]+273.15)**4)) # [W/m2K]
		newT=Ts[:,i]+Ksh*heat_flux*AV/rho/cp*(time[i+1]-time[i])
	Ts[:,i+1]=newT # assign last temperature
	
	return np.squeeze(Ts)

def Temperature_LumpedMass_Protected_const(Tg,time,AV,c_ins,rho_ins,l_ins,d_ins,n=1):
	""" Protected steel section temperature, according to section 4.2.5.2 EC;
	constant insulation properties

	Parameters
	----------	
	Tg : np.array(m,) or np.array(n,m)
		[C], adiabatic surface temperature (gas temperature)
		with n the number of fire curves, and m the number of timesteps
	time : np.array(m,)
		[s], timesteps for the calculation
	Av  : float or np.array(n,)
		[1/m], section factor for the unprotected section
	c_ins : float or np.array(n,)
		[J/kgK], insulation specific heat
	rho_ins : float or np.array (n,)
		[kg/m3], insulation density
	l_ins : float or np.array(n,)
		[W/mK], insulation conductivity
	d_ins : float or np.array(n,)
		[m], insulation thickness
	n   : float
		[-] number of simulations (explicitly transferred for ease of use)

	Returns
	-------
	Ts : np.array(n,m) or np.array(m,)
	 [C],  temperature for unprotected steel section

	Example use
	-----------
	import numpy as np
	import magnelPy.SFE as sfe
	from magnelPy.SFE.Capacity import EC3 as EC3
	from magnelPy.SFE.MaterialProperties import Steel as steel
	# parameters steel temperature calculation
	AV = 128 # [1/m]; specific surface section of the insulated steel member
	l_ins=0.2 # [W/mK]; insulation conductivity
	rho_ins=800 # [kg/m3]; insulation density
	c_ins=1000 # [J/kgK]; insulation specific heat
	d_ins=0.02 # [m]; insulation thickness
	# fire curve definition
	timelist=np.arange(0,360*60+0.5,0.5) #[s]; timesteps
	O = 0.04; qf=780 # parameters EPFC
	Tg=sfe.FireCurve.EuroCodeParametric(timelist/60, O, qf)
	# steel temperature calculation
	Ts=EC3.Temperature_LumpedMass_Protected_const(Tg,timelist,AV,c_ins,rho_ins,l_ins,d_ins)
	# output print
	times=[0,60,120,180,240,360] # [min]
	steps=np.array(times)*60*2
	for i,timepoint in enumerate(times):
		print("At {0:.0f} min, the temperature is {1:.0f} degrees C.".format(timepoint,Ts[steps[i]]))
	>>> At 0 min, the temperature is 20 degrees C.
	>>> At 60 min, the temperature is 448 degrees C.
	>>> At 120 min, the temperature is 619 degrees C.
	>>> At 180 min, the temperature is 521 degrees C.
	>>> At 240 min, the temperature is 274 degrees C.
	>>> At 360 min, the temperature is 54 degrees C.
	"""
	
	# hardcoded properties
	rho=steel.density_EC3() # [kg/m3] steel density (constant)
	
	# initialization
	m=len(time)
	if np.shape(Tg)!=(n,m):
		Tg=np.tile(Tg,(n,1))
	Ts = np.zeros((n,m))
	newT=20*np.ones(n) # [C] starting temperature

	# forward integration
	for i in np.arange(m-1):
		# assign temperature
		Ts[:,i]=newT
		dTg=Tg[:,i+1]-Tg[:,i] # [C]; temperature increase to next time step
		# material properties at elevated temperature
		cp=steel.cp_EC3(Ts[:,i]) # [J/kgK]; specific heat
		# steel temperature iteration
		phi=c_ins*rho_ins/(cp*rho)*d_ins*AV
		dTs=l_ins*AV/(d_ins*cp*rho)*(Tg[:,i]-Ts[:,i])/(1+phi/3)*(time[i+1]-time[i])-(np.exp(phi/10)-1)*dTg # [C], temperature increase
		delta_Ts=np.maximum(dTs,0,where=dTg>0) # [C], corrected temperature increase
		newT=Ts[:,i]+delta_Ts
	Ts[:,i+1]=newT # assign last temperature
	
	return np.squeeze(Ts)

def Temperature_LumpedMass_Protected_UserDef(Tg,time,AV,ins,d_ins,n=1):
	""" Protected steel section temperature, according to section 4.2.5.2 EC;
	UserDefined temperature-dependent insulation properties

	Parameters
	----------	
	Tg : np.array(m,) or np.array(n,m)
		[C], adiabatic surface temperature (gas temperature)
		with n the number of fire curves, and m the number of timesteps
	time : np.array(m,)
		[s], timesteps for the calculation
	Av  : float or np.array(n,)
		[1/m], section factor for the unprotected section
	ins : object or list of objects with len(list)=n
		object of UserDefined class SFE\\MaterialProperties\\UserDefined;
		contains call for conductivity, specific heat and density (l_ins [W/(m*K)], c_ins [J/(kg*K)], rho_ins [kg/m3])
	d_ins : float or np.array(n,)
		[m], insulation thickness
	n   : float
		[-] number of simulations (explicitly transfer for ease of use)

	Returns
	-------
	Ts : np.array(n,m) or np.array(m,)
	 [C],  temperature for unprotected steel section

	Example use
	-----------
	import numpy as np
	import magnelPy.SFE as sfe
	from magnelPy.SFE.Capacity import EC3 as EC3
	from magnelPy.SFE.MaterialProperties import Steel as steel
	from magnelPy.SFE.MaterialProperties import SFRM as sfrm
	import magnelPy.SFE.MaterialProperties as mat
	# parameters steel temperature calculation
	AV = 128 # [1/m]; specific surface section of the insulated steel member
	T=np.arange(20,1250,10) # [C]; temperatures for array-wise insulation material definition
	k=sfrm.ki(T) # [W/mK]; sfrm conductivity - used as UserDefined input
	c=sfrm.cp(T) # [kg/m3]; sfrm specific heat - used as UserDefined input
	r=sfrm.rho(T) #  [J/kgK]; sfrm density - used as UserDefined input
	d_ins=0.02 # [m]; insulation thickness
	ins=mat.UserDefined(T,k,c,r) # UserDefined material initialization
	# fire curve definition
	timelist=np.arange(0,360*60+0.5,0.5) #[s]; timesteps
	O = 0.04; qf=780 # parameters EPFC
	Tg=sfe.FireCurve.EuroCodeParametric(timelist/60, O, qf)
	# steel temperature calculation
	Ts=EC3.Temperature_LumpedMass_Protected_UserDef(Tg,timelist,AV,ins,d_ins)
	# output print
	times=[0,60,120,180,240,360] # [min]
	steps=np.array(times)*60*2
	for i,timepoint in enumerate(times):
		print("At {0:.0f} min, the temperature is {1:.0f} degrees C.".format(timepoint,Ts[steps[i]]))
	>>> At 0 min, the temperature is 20 degrees C.
	>>> At 60 min, the temperature is 431 degrees C.
	>>> At 120 min, the temperature is 630 degrees C.
	>>> At 180 min, the temperature is 542 degrees C.
	>>> At 240 min, the temperature is 373 degrees C.
	>>> At 360 min, the temperature is 168 degrees C.
	"""
	
	# hardcoded properties
	rho=steel.density_EC3() # [kg/m3] steel density (constant)
	
	# initialization
	m=len(time)
	if np.shape(Tg)!=(n,m):
		Tg=np.tile(Tg,(n,1))
	Ts = np.zeros((n,m))
	newT=20*np.ones(n) # [C] starting temperature
	if type(ins)!=list: ins=[ins]
	c_ins=np.zeros(n)
	rho_ins=np.zeros(n)
	l_ins=np.zeros(n)

	# forward integration
	for i in np.arange(m-1):
		# assign temperature
		Ts[:,i]=newT
		dTg=Tg[:,i+1]-Tg[:,i] # [C]; temperature increase to next time step
		# material properties at elevated temperature
		cp=steel.cp_EC3(Ts[:,i]) # [J/kgK]; specific heat
		avg_temp = (Tg[:,i] + Ts[:,i])/2 # average insulation temperature
		for j,ins_mat in enumerate(ins):
			c_ins[j] = ins_mat.cp(avg_temp[j]); rho_ins[j] = ins_mat.rho(avg_temp[j]); l_ins[j] = ins_mat.k(avg_temp[j])
		# steel temperature iteration
		phi=c_ins*rho_ins/(cp*rho)*d_ins*AV
		dTs=l_ins*AV/(d_ins*cp*rho)*(Tg[:,i]-Ts[:,i])/(1+phi/3)*(time[i+1]-time[i])-(np.exp(phi/10)-1)*dTg # [C], temperature increase
		delta_Ts=np.maximum(dTs,0,where=dTg>0) # [C], corrected temperature increase
		newT=Ts[:,i]+delta_Ts
	Ts[:,i+1]=newT # assign last temperature
	
	return np.squeeze(Ts)

#########
## AUX ##
#########